//----------------------------------------------------------------------------
// The Famous DirectX Demo
//
// reference dlls (various include paths, add yours if different) 
//
// THIS DEMO IS A SCRIPT - USE F# INTERACTIVE.  DO NOT COMPILE & RUN
//
//----------------------------------------------------------------------------

// adjust these as needed for your latest installed version of DirectX

#I @"C:\WINDOWS\Microsoft.NET\DirectX for Managed Code\1.0.2902.0" ;; 
#I @"C:\WINDOWS\Microsoft.NET\DirectX for Managed Code\1.0.2903.0" ;; 
#r @"Microsoft.DirectX.dll";;
#r @"Microsoft.DirectX.Direct3D.dll" ;;
#r @"Microsoft.DirectX.Direct3Dx.dll" ;;

// !!!!! Use this line when using F# Interactive:
//
#load @"dxlib.fs";;


open System
open Drawing
open System.Threading  
open System.Windows.Forms
open Microsoft.DirectX
open Microsoft.DirectX.Direct3D
open Idioms
open List
open Sample.DirectX.Basics

//----------------------------------------------------------------------------
// DirectX form
//
// Create an enclosing form which owns its own painting.
//----------------------------------------------------------------------------

let form = new SmoothForm()
do form.Text <- "F# surface plot"
do form.ClientSize <- new Size(400,300)
do form.Visible <- true
do form.TopMost <- true

//----------------------------------------------------------------------------
// DirectX device
//----------------------------------------------------------------------------

let presentParams = defaultPresentParams()

// Create device   
let device =
  (new Device(0, DeviceType.Hardware, form,
                 CreateFlags.SoftwareVertexProcessing, 
                 [| presentParams |]))

//----------------------------------------------------------------------------
// The drawScene event fires each time we draw the scene.
// The event is fired within the context of a BeginScene/EndScene.
// It gets passed the current time, so all recipients see
// a consistent view of time.
//----------------------------------------------------------------------------

let (doDrawScene,drawScene) : event<time> = IEvent.create()

//----------------------------------------------------------------------------
// The render function.  Raises the drawScene event
// after checking the device is ready for use.
//----------------------------------------------------------------------------

do form.Paint.Add(fun _ -> 
   let t = now() in 
   doRender device presentParams form (fun () -> doDrawScene t); 
   form.Invalidate())

do device.DeviceReset.Add(fun _ -> doInitialise device form)

do doInitialise device form;;

//----------------------------------------------------------------------------
// XY grid and axis box
//----------------------------------------------------------------------------
  
let drawAxis (device:Device) =
  device.RenderState.CullMode <- Cull.None;
  drawLineList planePts device;
  drawLineList boxPts device

do drawScene.Add(fun _ -> drawAxis device)

//----------------------------------------------------------------------------
// DirectX setup lighting
//----------------------------------------------------------------------------

do drawScene.Add(fun time -> setupLights device)

//----------------------------------------------------------------------------
// setView
//----------------------------------------------------------------------------
    
// some interactive state (interactive state would eventually be subsumed into state of a chart object)   
let view = 
   { ypr   = Matrix.RotationYawPitchRoll(0.0f,0.0f,0.0f);
     focus = scale 0.5f (vX + vY + vZ);
     zoom  = 4.0 }

do drawScene.Add(fun time -> setView view device)

//----------------------------------------------------------------------------
// Mouse events
//----------------------------------------------------------------------------

let mouseEvent = mkMouseTracker form
do mouseEvent.Add(move view)

//----------------------------------------------------------------------------
// mesh and gridLines
//----------------------------------------------------------------------------

let normi n i = float i / float (n-1)
let grid (m,n) = mkMesh normi normi (m,n)

let mesh = ref (grid (20,20))

let dims () = meshDims !mesh

let gridLine i j  = [i,j;i+1,j;  i,j;i,j+1]
let gridLines n m = mapR 0 (n-2) (fun i -> mapR 0 (m-2) (gridLine i)) |> concat |> concat

let drawSurf f (device:Device) =
  let m,n = meshDims !mesh in 
  let data = Array.init (n*m) (fun k ->
                                 let i,j = k2ij !mesh k in
                                 let x,y = meshGet !mesh (i,j) in
                                 let z = f (x,y) in // single precision f 
                                 (x,y,z))
  in
  let strips = triangleRows n m |> map (map (blendPlace !mesh data)) in
  List.iter (fun strip -> drawTriangeStrip strip device) strips;
  let m,n = meshDims !mesh in 
  let lines = gridLines n m in
  let lines = lines |> map (colorPlace !mesh data Color.Black) in
  drawLineList lines device

//----------------------------------------------------------------------------
// state: evaluation function
//----------------------------------------------------------------------------

/// function to display
let ff    = ref (fun (t:float) (x,y) -> x * (1.0 - y))
/// Z-range
let range = ref (0.0,1.0)

let eval t (x,y) =
  let z = !ff t (x,y) in
  let a,b = !range in // and scale w.r.t. range ... 
  let res = if b - a = 0.0 then 0.0 else (z-a) / (b-a) in
  res

do drawScene.Add(fun time -> drawSurf (eval time) device);;

//----------------------------------------------------------------------------
// PART 2 - change the function
//----------------------------------------------------------------------------

do ff := fun t (x,y) -> sqr (x - 0.5) * sqr (y - 0.5) * 16.0;;
do ff := fun t (x,y) -> 0.5 * sin(x * 4.5 + t / 2.0) * cos(y * 8.0) * x + 0.5;;
do range := (-1.0,1.0)
do range := (0.0,1.0)

let ripple t (x,y) =
   let x,y = x - 0.5,y - 0.5 in
   let r = sqrt (x*x + y*y) in
   exp(-4.0 * r) * sin(6.0 * 3.14 * r + t) + 0.5
do ff := ripple


//----------------------------------------------------------------------------
// PART 3 - change the grid
//----------------------------------------------------------------------------
    

do mesh := grid (50,50);;
do mesh := grid (20,20);;

let cgrid1 fd m = 
  let dxs = mapR 0 (m-2) (normi (m-1) >> fd) in
  let sumx = fold_right (+) dxs 0.0 in
  let _,xs = fold_right (fun dx (sum,res) -> (sum+dx),(sum/sumx :: res)) dxs (0.0,[]) in
  let arr = Array.of_list (1.0 :: xs) in
  fun i -> arr.(i)
let cgrid fdensity (m,n) = mkMesh (cgrid1 fdensity) (cgrid1 fdensity) (m,n)

do mesh := cgrid (fun x -> sqr ( (x - 0.5) * 2.0 ) + 0.2) (20,20);;

// First, set up a bowl shaped area 
do range := (0.0,2.0)
let surfaceA t (x,y) = let f phi u = ((1.0 + cos(2.0 * pi * u + phi )) / 2.0) in
                       f t x * f 0.0 y + 1.0;;

ff := fun t (x,y) -> surfaceA 0.0 (x,y);;

//----------------------------------------------------------------------------
// PART 4 - Motion on a surface
//----------------------------------------------------------------------------    

let gravity = ref (new Vector3(0.0f,0.0f,-5.0f))

let surfacePoint f (x,y) =
  let z = f (x,y) in
  new Vector3(x,y,z)
  
let surfaceNormal f (x,y) =
  let dx,dy = 0.01f,0.01f in
  let pA    = surfacePoint f (x,y) in
  let pA_dx = surfacePoint f (x+dx,y) - pA in
  let pA_dy = surfacePoint f (x,y+dy) - pA in
  normalize (cross pA_dx pA_dy)

type ball = Ball of Vector3 * Vector3 

let radiusA = 0.010f 
let radiusB = 0.005f     

let stepMotion f dt (Ball (position,velocity)) =
  let nHat     = surfaceNormal f (position.X,position.Y) in // surface plane normal   
  let acc      = planeProject nHat !gravity in               // acceleration in plane 
  let velocity = planeProject nHat velocity in              // velocity     in plane 
  let position = position + Vector3.Scale(velocity,dt) in   // iterate 
  let velocity = velocity + Vector3.Scale(acc     ,dt) in   // iterate 
  let bounce (p,v) =                                        
    if p < 0.0f + radiusA then (2.0f * (0.0f + radiusA) - p,-v) else  
    if p > 1.0f - radiusA then (2.0f * (1.0f - radiusA) - p,-v) else
    (p,v) : float32 * float32
  in
  let px,vx = bounce (position.X,velocity.X) in             // bounce X edges 
  let py,vy = bounce (position.Y,velocity.Y) in             // bounce Y edges 
  let position = surfacePoint f (px,py) in                  // keep to surface 
  let velocity = new Vector3 (vx,vy,velocity.Z) in
  let velocity = planeProject nHat velocity in              // velocity in plane     
  Ball (position,velocity)

let position = surfacePoint (singleFun2 (eval (now()))) (0.2f,0.8f)
let velocity = surfacePoint (singleFun2 (eval (now()))) (0.1f,0.4f)
let ballP = Ball (position,velocity)

//----------------------------------------------------------------------------
// PART 5 - Motion on a surface rendered
//----------------------------------------------------------------------------

let drawPlaneArrow device n p dir = 
  let dir2 = normalize (cross n dir) in
  drawLineList (
    map point [p + scale 0.15f dir; p;
               p + scale 0.15f dir; p + scale 0.10f dir + scale 0.02f dir2;
               p + scale 0.15f dir; p + scale 0.10f dir - scale 0.02f dir2]) device

let drawBall device t (Ball (p,v)) =
  let n    = surfaceNormal (singleFun2 (eval t)) (p.X,p.Y) in
  let p0   = new Vector3(p.X,p.Y,0.0f) in
  let pV   = new Vector3(v.X,v.Y,0.0f) in                 // unit velocity XY-projection 
  let pVxZ = Vector3.Cross(pV,vZ) in                        // and it's XY-perpendicular 
  drawLineList (map (colorPoint Color.Gray) [p0;p]) device; // vertical line 
  drawPlaneArrow device vZ           p0 pV;                 // velocity arrow on floor 
  drawPlaneArrow device (cross n vX) p  (scale 0.8f n);     // normal arrow at point     
  device.Transform.World <- (let m = Matrix.LookAtLH(p + scale radiusB n,p+n,vX) in Matrix.Invert(m));
  let mesh = Mesh.Torus(device,radiusB,radiusA,20,20) in
  mesh.ComputeNormals();
  //let mesh = Mesh.Sphere(device,radiusA,20,20) in
  mesh.DrawSubset(0);
  mesh.Dispose();
  device.Transform.World <- Matrix.Identity;;


//----------------------------------------------------------------------------
// Motion on a surface example
//----------------------------------------------------------------------------

//do timer.Interval <- 50 // ms = frame rate 

let activeList = ref [] : ball list ref
let addActiveList ball = activeList := (ball :: !activeList)  

let simulate t = 
  let dt = 0.005f in 
  activeList := List.map (stepMotion (singleFun2 (eval t)) dt) !activeList

do drawScene |> IEvent.listen simulate
do drawScene.Add(fun t -> List.iter (drawBall device t) !activeList)


// Second, add a ball 
let ballA = Ball (new Vector3(0.1f,0.1f,0.1f),new Vector3(0.6f,0.5f,0.0f))
do  activeList := []
do  addActiveList ballA 

// Third, add a train of balls to a moving floor. Vary rate... 

let spawn (f : unit -> unit) = ThreadPool.QueueUserWorkItem(fun _ -> f ()) |> ignore;;
do  activeList := [];;
do  spawn(fun () -> iterR 0 6 (fun i -> addActiveList ballA; Thread.Sleep(300)));;

let rate = 0.25 
do ff := fun t (x,y) -> surfaceA (rate * t) (x,y)
do ff := ripple;;

//----------------------------------------------------------------------------
// PART 6 - Make the basic scene drawing framework into a component
//----------------------------------------------------------------------------

do form.Close();;
  
type DisplayControl = class 

  inherit UserControl
  // The drawScene event fires each time we draw the scene.
  val drawSceneE : event<time>
  val mutable device : Device 
  
  member x.DrawScene = snd x.drawSceneE
  member x.Device = x.device
  
  new() as form = 
    let drawSceneE = IEvent.create() in
    { inherit UserControl(); 
      drawSceneE = drawSceneE;
      device = null } 
   then
    let doDrawScene,drawScene = form.drawSceneE in 
    form.SetStyle(Enum.combine [ControlStyles.AllPaintingInWmPaint; ControlStyles.Opaque], true);
    let presentParams = defaultPresentParams() in 
    form.CreateControl();
    // Create device   
    let device = 
      (new Device(0, DeviceType.Hardware, form,
                     CreateFlags.SoftwareVertexProcessing, 
                     [| presentParams |])) in
    form.device <- device;
   
    // We must reset the device if it is rudely taken from us
    let doReset () = form.device.Reset([|presentParams|]) in 

    // setup     
    form.Paint.Add(fun _ -> 
      doRender form.device presentParams form (fun () -> doDrawScene (now())); 
      Application.DoEvents(); 
      form.Invalidate());
    form.device.DeviceReset.Add(fun _ -> doInitialise form.device form);
    doInitialise form.device form;

    let mouseEvent = mkMouseTracker form in 
    let view =  
      { ypr   = Matrix.RotationYawPitchRoll(0.0f,0.0f,0.0f);
        focus = scale 0.5f (vX + vY + vZ);
        zoom  = 4.0 } in
    
    mouseEvent.Add(move view);
    drawScene.Add(fun t -> setView view form.device);
 end


let form2 = new Form()
let control2 = new DisplayControl()
do form2.TopMost <- true
do form2.Visible <- true
do control2.Dock <- DockStyle.Fill
do form2.Controls.Add(control2)
do control2.DrawScene.Add(fun t -> drawAxis control2.Device)
do control2.DrawScene.Add(fun t -> drawSurf (eval t) control2.Device)
do control2.DrawScene.Add(fun _ -> setupLights control2.Device);;

